// $Id: cube.c 12 2007-03-03 20:23:17Z salival $
/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <stdint.h>
#include <avr/io.h>
#include "cube.h"

#if CUBE_SIZE > 15
#       error "3375 LEDs should be enough for everyone"
#endif //CUBE_SIZE

struct {
     uint8_t display_ram[CUBE_DISPLAY_RAM_SIZE];
     uint8_t next_step[CUBE_SHIFTREGISTER_N];
     uint8_t layer_counter;
     uint8_t pwm_counter;
} cube;

void cube_init(void)
{
    register uint8_t i;

    for (i=0; i<CUBE_SHIFTREGISTER_N; i++)
        cube.next_step[i] = 0;

    cube.layer_counter = 0;
    cube.pwm_counter = 0;
    cube_clear();
}

void cube_clear(void)
{
    register uint16_t i;
    for (i=0; i<CUBE_DISPLAY_RAM_SIZE; i++)
        cube.display_ram[i] = 0x00;
}

void cube_calc_next_step(void)
{
    register uint8_t shiftregister_counter = CUBE_SHIFTREGISTER_N;
    register uint8_t led_counter = CUBE_LED_LAYER_N;
    register uint8_t pwm_counter = cube.pwm_counter;
    register uint8_t layer_counter = cube.layer_counter;
    register uint8_t high_nibble;
    register uint8_t ram_value;
    register uint8_t shiftregister_value;
    register uint8_t i;
    uint8_t led_flag = 1;
    uint8_t* pointer;
    uint16_t offset;

    if ( (++pwm_counter) >= CUBE_PWM_STEPS)
    {
        pwm_counter = 0;
        if ( (++layer_counter) >= CUBE_SIZE)
        {
            layer_counter = 0;
        }
    }
    cube.layer_counter = layer_counter;
    cube.pwm_counter = pwm_counter;

    offset = layer_counter*CUBE_LED_LAYER_N;
    high_nibble = offset%2;
    offset /= 2;
    pointer = &cube.display_ram[offset];
    ram_value = *(pointer++);

    while (shiftregister_counter--)
    {
        shiftregister_value = 0;

        for (i=0; i<8; i++)
        {
            if (led_flag)
            {
                led_flag = (--led_counter) ? 1 : 0;
                if (high_nibble)
                {

                    if ( ((ram_value>>4)) > pwm_counter )
                    {
                        shiftregister_value |= (1<<i);
                    }
                    ram_value = *(pointer++);
                }
                else
                {

                    if ( ((ram_value & 0x0f)) > pwm_counter )
                    {
                        shiftregister_value |= (1<<i);
                    }
                }

                high_nibble ^= 1;
            }
            else
            {
                if (!layer_counter--)
                {
                    shiftregister_value |= (1<<i);
                    break;
                }
            }
        }

        cube.next_step[shiftregister_counter] = shiftregister_value;
    }
}

uint8_t* cube_get_step_pointer(void)
{
    return &cube.next_step[0];
}

uint8_t* cube_get_display_pointer(void)
{
    return &cube.display_ram[0];
}

void cube_set(uint8_t x, uint8_t y, uint8_t z, uint8_t value)
{
    uint16_t led_position = x + y*CUBE_LED_LAYER_N + z*CUBE_SIZE;
    uint16_t ram_position = led_position/2;
    register uint8_t tmp = cube.display_ram[ram_position];

    //high nibble
    if (led_position % 2)
    {
        tmp &= 0x0f;
        tmp |= value<<4;
    }
    //low nibble
    else
    {
        tmp &= 0xf0;
        tmp |= (value & 0x0f);
    }

    cube.display_ram[ram_position] = tmp;
}

